import tempfile
import uuid
import zoneinfo
from datetime import date, datetime

import pandas as pd

from ..kawa_base_e2e_test import KawaBaseTest


def utc(year, month, day, hour, minute, second):
    return datetime(year, month, day, hour, minute, second, tzinfo=zoneinfo.ZoneInfo('UTC'))


class TestComputationSample(KawaBaseTest):

    @classmethod
    def setUpClass(cls):
        unique_id = 'resource_{}'.format(uuid.uuid4())
        print(tempfile.gettempdir())
        KawaBaseTest.setUpClass()
        cls.data, df = cls.df_for_tests()
        cls.sheet_name = unique_id
        loader = cls.kawa.new_data_loader(df=df, datasource_name=unique_id)
        loader.create_datasource()
        loader.load_data(reset_before_insert=True, create_sheet=True)

    def test_date_year_and_month_sampling(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .sample(sampler='YEAR_AND_MONTH')
                       .group_by('date')
                       .agg(self.kawa.col('decimal').sum())
                       .no_limit()
                       .compute())

        # then
        self.assertListEqual(list(computed_df['grouping(0)']), ['2023-01'])
        self.assertListEqual(list(computed_df['decimal']), [1.1 + 3.1 + 2.1 + 4.1 + 5.1])

    def test_date_year_and_quarter_sampling(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .sample(sampler='YEAR_AND_QUARTER')
                       .group_by('date')
                       .agg(self.kawa.col('decimal').sum())
                       .no_limit()
                       .compute())

        # then
        self.assertListEqual(list(computed_df['grouping(0)']), ['2023-1'])
        self.assertListEqual(list(computed_df['decimal']), [1.1 + 3.1 + 2.1 + 4.1 + 5.1])

    def test_date_year_and_semester_sampling(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .sample(sampler='YEAR_AND_SEMESTER')
                       .group_by('date')
                       .agg(self.kawa.col('decimal').sum())
                       .no_limit()
                       .compute())

        # then
        self.assertListEqual(list(computed_df['grouping(0)']), ['2023-1'])
        self.assertListEqual(list(computed_df['decimal']), [1.1 + 3.1 + 2.1 + 4.1 + 5.1])

    def test_date_day_of_year_sampling(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .sample(sampler='DAY_OF_YEAR')
                       .group_by('date')
                       .agg(self.kawa.col('decimal').sum())
                       .no_limit()
                       .compute())

        # then
        self.assertListEqual(sorted(list(computed_df['grouping(0)'])), sorted(['001', '002']))
        self.assertListEqual(sorted(list(computed_df['decimal'])), sorted([1.1 + 3.1, 2.1 + 4.1 + 5.1]))

    def test_date_year_sampling(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .sample(sampler='YEAR')
                       .group_by('date')
                       .agg(self.kawa.col('decimal').sum())
                       .no_limit()
                       .compute())

        # then
        self.assertListEqual(list(computed_df['grouping(0)']), ['2023'])
        self.assertListEqual(list(computed_df['decimal']), [1.1 + 3.1 + 2.1 + 4.1 + 5.1])

    def test_date_month_sampling(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .sample(sampler='MONTH')
                       .group_by('date')
                       .agg(self.kawa.col('decimal').sum())
                       .no_limit()
                       .compute())

        # then
        self.assertListEqual(list(computed_df['grouping(0)']), ['01'])
        self.assertListEqual(list(computed_df['decimal']), [1.1 + 3.1 + 2.1 + 4.1 + 5.1])

    def test_date_week_sampling(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .sample(sampler='WEEK')
                       .group_by('date')
                       .agg(self.kawa.col('decimal').sum())
                       .no_limit()
                       .compute())

        # then
        self.assertListEqual(list(computed_df['grouping(0)']), ['01', '52'])
        self.assertListEqual(list(computed_df['decimal']), [+ 2.1 + 4.1 + 5.1, 1.1 + 3.1])

    def test_date_time_year_sampling(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .sample(sampler='YEAR')
                       .group_by('date_time')
                       .agg(self.kawa.col('decimal').sum())
                       .no_limit()
                       .compute())

        # then
        self.assertListEqual(list(computed_df['grouping(0)']), ['2023'])
        self.assertListEqual(list(computed_df['decimal']), [1.1 + 3.1 + 2.1 + 4.1 + 5.1])

    def test_date_time_hour_sampling(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name, force_tz='UTC')
                       .sample(sampler='HOUR')
                       .group_by('date_time')
                       .agg(self.kawa.col('decimal').sum())
                       .no_limit()
                       .compute())

        # then
        self.assertListEqual(list(computed_df['grouping(0)']), [utc(2023, 1, 1, 23, 0, 0)])
        self.assertListEqual(list(computed_df['decimal']), [1.1 + 3.1 + 2.1 + 4.1 + 5.1])

    def test_date_time_fifteen_minutes_sampling(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name, force_tz='UTC')
                       .sample(sampler='FIFTEEN_MINUTES')
                       .group_by('date_time')
                       .agg(self.kawa.col('decimal').sum())
                       .no_limit()
                       .compute())

        # then
        self.assertListEqual(list(computed_df['grouping(0)']), [utc(2023, 1, 1, 23, 15, 0)])
        self.assertListEqual(list(computed_df['decimal']), [1.1 + 3.1 + 2.1 + 4.1 + 5.1])

    def test_date_time_minute_sampling(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name, force_tz='UTC')
                       .sample(sampler='MINUTE')
                       .group_by('date_time')
                       .agg(self.kawa.col('decimal').sum())
                       .no_limit()
                       .compute())

        # then
        self.assertListEqual(list(computed_df['grouping(0)']), [utc(2023, 1, 1, 23, 23, 0)])
        self.assertListEqual(list(computed_df['decimal']), [1.1 + 3.1 + 2.1 + 4.1 + 5.1])

    def test_date_time_thirty_seconds_sampling(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name, force_tz='UTC')
                       .sample(sampler='THIRTY_SECONDS')
                       .group_by('date_time')
                       .agg(self.kawa.col('decimal').sum())
                       .no_limit()
                       .compute())

        # then
        self.assertListEqual(list(computed_df['grouping(0)']), [utc(2023, 1, 1, 23, 23, 0)])
        self.assertListEqual(list(computed_df['decimal']), [1.1 + 3.1 + 2.1 + 4.1 + 5.1])

    @staticmethod
    def df_for_tests():
        data = {
            'date': [
                date(2023, 1, 1),
                date(2023, 1, 2),
                date(2023, 1, 1),
                date(2023, 1, 2),
                date(2023, 1, 2)
            ],
            'date_time': [
                utc(2023, 1, 1, 23, 23, 2),
                utc(2023, 1, 1, 23, 23, 4),
                utc(2023, 1, 1, 23, 23, 2),
                utc(2023, 1, 1, 23, 23, 2),
                utc(2023, 1, 1, 23, 23, 2),
            ],
            'text': ['value11', 'value12', 'value21', 'value22', 'value23'],
            'integer': [1, 2, 3, 4, 5],
            'decimal': [1.1, 2.1, 3.1, 4.1, 5.1],
            'boolean': [True, False, True, True, False],
        }

        data_for_df = []
        for i in range(5):
            data_for_df.append({
                'date': data['date'][i],
                'date_time': data['date_time'][i],
                'boolean': data['boolean'][i],
                'text': data['text'][i],
                'integer': data['integer'][i],
                'decimal': data['decimal'][i],
            })

        return data, pd.DataFrame(data_for_df)



import typing
import pyarrow
from datetime import date, datetime
import datetime as dt
from tzlocal import get_localzone
import type_enforced

from ..client.computation_nodes import TreeNode


@type_enforced.Enforcer(enabled=True)
class KawaColumn:

    def __init__(self,
                 column_name: typing.Optional[str],
                 column_regexp: typing.Optional[str] = None,
                 column_alias: typing.Optional[str] = None,
                 indicator_columns_only: bool = False,
                 xml_syntactic_tree: typing.Optional[str] = None,
                 default_sheet_columns_only: bool = False):
        self._aggregation_method = None
        self._column_name = column_name
        self._column_alias = column_alias
        self._column_regexp = column_regexp
        self._indicator_columns_only = indicator_columns_only
        self._xml_syntactic_tree = xml_syntactic_tree
        self._default_sheet_columns_only = default_sheet_columns_only

    def to_dict(self):
        return {
            'columnName': self.column_name(),
            'columnAlias': self.column_alias(),
            'aggregationMethod': self.aggregation_method(),
            'columnRegexp': self.column_regexp(),
            'indicatorColumnsOnly': self.indicator_columns_only(),
            'sheetDefaultColumnsOnly': self.default_sheet_columns_only(),
            'xmlSyntacticTree': self.xml_syntactic_tree(),
        }

    def xml_syntactic_tree(self):
        return self._xml_syntactic_tree

    def indicator_columns_only(self):
        return self._indicator_columns_only

    def default_sheet_columns_only(self):
        return self._default_sheet_columns_only

    def column_regexp(self):
        return self._column_regexp

    def column_name(self):
        return self._column_name

    def aggregation_method(self):
        return self._aggregation_method

    def column_alias(self):
        return self._column_alias

    def aggregate(self, aggregation_method: str):
        self._aggregation_method = aggregation_method.upper()
        return self

    def filter(self):
        return KawaFilter(column=self)

    def alias(self, alias: str):
        self._column_alias = alias.strip()
        return self

    def filter_with_list_of_values(self, values: list[str]):
        return self.filter().in_list(values)

    # Filters
    def yoy_ytd(self):
        return self.filter().yoy_ytd()

    def date_range(self,
                   from_inclusive: typing.Optional[date] = None,
                   to_inclusive: typing.Optional[date] = None):
        return self.filter().date_range(from_inclusive, to_inclusive)

    def datetime_range(self,
                       from_inclusive: typing.Optional[datetime] = None,
                       to_inclusive: typing.Optional[datetime] = None):
        return self.filter().datetime_range(from_inclusive, to_inclusive)

    def time_range(self,
                   from_inclusive: typing.Optional[dt.time] = None,
                   from_exclusive: typing.Optional[dt.time] = None,
                   to_inclusive: typing.Optional[dt.time] = None,
                   to_exclusive: typing.Optional[dt.time] = None):
        return self.filter().time_range(
            from_inclusive=from_inclusive,
            from_exclusive=from_exclusive,
            to_inclusive=to_inclusive,
            to_exclusive=to_exclusive
        )

    def empty(self):
        return self.filter().empty()

    def not_empty(self):
        return self.filter().not_empty()

    def in_list(self, *args):
        # Note: This argument introspection is done to maintain backward compatibility
        # With pre 0.18beta7 versions
        # (commit 7b1ccb11)
        # Two ways to call this:
        # - With a tuple: K.col('text').in_list('this', 'is', 'foo')
        # - With a list:  K.col('text').in_list(['this', 'is', 'foo'])
        list_of_values = []
        if args is None:
            return self.filter().in_list([])
        for arg in args:
            if isinstance(arg, list):
                list_of_values.extend([str(v) for v in arg])
            else:
                list_of_values.append(str(arg))

        return self.filter().in_list(list_of_values)

    def starts_with(self, value: str):
        return self.filter().starts_with(value)

    def ends_with(self, value: str):
        return self.filter().ends_with(value)

    def contains(self, value: str):
        return self.filter().contains(value)

    def does_not_contain(self, value: str):
        return self.filter().does_not_contain(value)

    def does_not_start_with(self, value: str):
        return self.filter().does_not_start_with(value)

    def does_not_end_with(self, value: str):
        return self.filter().does_not_end_with(value)

    def gt(self, value: [float, int]):
        return self.filter().gt(value)

    def lt(self, value: [float, int]):
        return self.filter().lt(value)

    def gte(self, value: [float, int]):
        return self.filter().gte(value)

    def lte(self, value: [float, int]):
        return self.filter().lte(value)

    def eq(self, value: [float, int, str]):
        return self.filter().eq(value)

    def ne(self, value: [float, int, str]):
        return self.filter().ne(value)

    # Aggregations
    def first(self):
        return self.aggregate('first')

    def identical(self):
        return self.aggregate('identical')

    def identical_ignore_empty(self):
        return self.aggregate('identical_ignore_empty')

    def count(self):
        return self.aggregate('count')

    def count_unique(self):
        return self.aggregate('count_unique')

    def percent_filled(self):
        return self.aggregate('percent_filled')

    def percent_empty(self):
        return self.aggregate('percent_empty')

    def count_empty(self):
        return self.aggregate('count_empty')

    def sum(self):
        return self.aggregate('sum')

    def avg(self):
        return self.aggregate('avg')

    def median(self):
        return self.aggregate('median')

    def min(self):
        return self.aggregate('min')

    def max(self):
        return self.aggregate('max')

    def min_abs(self):
        return self.aggregate('min_abs')

    def max_abs(self):
        return self.aggregate('max_abs')

    def var_sample(self):
        return self.aggregate('var_sample')

    def var_pop(self):
        return self.aggregate('var_pop')

    def std_dev_sample(self):
        return self.aggregate('std_dev_sample')

    def std_dev_pop(self):
        return self.aggregate('std_dev_pop')

    def lowest_decile(self):
        return self.aggregate('lowest_decile')

    def lowest_quartile(self):
        return self.aggregate('lowest_quartile')

    def highest_decile(self):
        return self.aggregate('highest_decile')

    def highest_quartile(self):
        return self.aggregate('highest_quartile')


@type_enforced.Enforcer(enabled=True)
class KawaFilter:

    def __init__(self,
                 column: typing.Optional[KawaColumn] = None,
                 indicator_id: typing.Optional[str] = None):
        self._column = column
        self._indicator_id = indicator_id
        if bool(column) == bool(indicator_id):
            raise AssertionError('One of column or indicator_id must be specified')

        self._exclude = False
        self._clauses = []

    def exclude(self):
        self._exclude = True
        return self

    def _add_clause(self, operator: str, value=None):
        self._clauses.append({
            'arguments': {'value': value} if value is not None else {},
            'operation': operator
        })
        return self

    def _add_clause_for_date_filter(self,
                                    from_days: typing.Optional[int],
                                    to_days: typing.Optional[int]):
        self._clauses.append({
            'arguments': {
                'dateFrom': from_days,
                'dateTo': to_days
            }
        })
        return self

    def _add_clause_for_date_time_filter(self,
                                         instant_milliseconds_from: typing.Optional[int],
                                         instant_milliseconds_to: typing.Optional[int]):
        self._clauses.append({
            'arguments': {
                'instantFrom': int(instant_milliseconds_from / 1000) if instant_milliseconds_from else None,
                'instantTo': int(instant_milliseconds_to / 1000) if instant_milliseconds_to else None,
                'instantMillisecondsFrom': instant_milliseconds_from,
                'instantMillisecondsTo': instant_milliseconds_to
            }
        })
        return self

    def _add_clause_for_time_filter(self,
                                    time_from_inclusive: typing.Optional[str],
                                    time_from_exclusive: typing.Optional[str],
                                    time_to_inclusive: typing.Optional[str],
                                    time_to_exclusive: typing.Optional[str]):
        self._clauses.append({
            'arguments': {
                'timeFromInclusive': time_from_inclusive,
                'timeFromExclusive': time_from_exclusive,
                'timeToInclusive': time_to_inclusive,
                'timeToExclusive': time_to_exclusive,
            }
        })
        return self

    def to_dict(self):
        d = {
            'exclude': self._exclude,
            'clauses': self._clauses,
        }

        if self._column:
            d['column'] = self._column.to_dict()
        if self._indicator_id:
            d['indicatorId'] = self._indicator_id

        return d

    # Empty / Not empty
    def not_empty(self):
        return self._add_clause('not_empty')

    def empty(self):
        return self._add_clause('empty')

    # Text clauses
    def in_list(self, values: list[str]):
        return self._add_clause('in_list', values)

    def contains(self, value: str):
        return self._add_clause('contains', value)

    def ends_with(self, value: str):
        return self._add_clause('ends_with', value)

    def starts_with(self, value: str):
        return self._add_clause('starts_with', value)

    def does_not_contain(self, value: str):
        return self._add_clause('does_not_contain', value)

    def does_not_end_with(self, value: str):
        return self._add_clause('does_not_end_with', value)

    def does_not_start_with(self, value: str):
        return self._add_clause('does_not_start_with', value)

    # Temporal filters
    def weekdays_only(self):
        if self._clauses and self._clauses[0]['arguments']:
            self._clauses[0]['arguments']['keepWeekDaysOnly'] = True
        return self

    def yoy_ytd(self):
        self._clauses.append({'arguments': {'specialMode': 'YOY_YTD'}})
        return self

    def date_range(self,
                   from_inclusive: typing.Optional[date] = None,
                   to_inclusive: typing.Optional[date] = None):
        from_days = (from_inclusive - date(1970, 1, 1)).days if from_inclusive else None
        to_days = (to_inclusive - date(1970, 1, 1)).days if to_inclusive else None
        return self._add_clause_for_date_filter(
            from_days=from_days,
            to_days=to_days)

    def datetime_range(self,
                       from_inclusive: typing.Optional[datetime] = None,
                       to_inclusive: typing.Optional[datetime] = None):
        from_milliseconds = int(from_inclusive.timestamp() * 1000) if from_inclusive else None
        to_milliseconds = int(to_inclusive.timestamp() * 1000) if to_inclusive else None
        return self._add_clause_for_date_time_filter(
            instant_milliseconds_from=from_milliseconds,
            instant_milliseconds_to=to_milliseconds)

    def time_range(self,
                   from_inclusive: typing.Optional[dt.time] = None,
                   from_exclusive: typing.Optional[dt.time] = None,
                   to_inclusive: typing.Optional[dt.time] = None,
                   to_exclusive: typing.Optional[dt.time] = None):

        if from_inclusive and from_exclusive:
            raise AssertionError('Both exclusive and inclusive from are defined')

        if to_inclusive and to_exclusive:
            raise AssertionError('Both exclusive and inclusive to are defined')

        return self._add_clause_for_time_filter(
            time_from_inclusive=str(from_inclusive) if from_inclusive else None,
            time_from_exclusive=str(from_exclusive) if from_exclusive else None,
            time_to_inclusive=str(to_inclusive) if to_inclusive else None,
            time_to_exclusive=str(to_exclusive) if to_exclusive else None,
        )

    # Numeric clauses
    def lt(self, value: [int, float]):
        return self._add_clause('lt', value)

    def lte(self, value: [int, float]):
        return self._add_clause('lte', value)

    def gt(self, value: [int, float]):
        return self._add_clause('gt', value)

    def gte(self, value: [int, float]):
        return self._add_clause('gte', value)

    def eq(self, value: [int, float, str]):
        return self._add_clause('eq', value)

    def ne(self, value: [int, float, str]):
        return self._add_clause('ne', value)


@type_enforced.Enforcer(enabled=True)
class KawaLazyQuery:

    def __init__(self,
                 kawa_client,
                 sheet_name: typing.Optional[str] = None,
                 force_tz: typing.Optional[str] = None,
                 no_output: bool = False):
        self._k = kawa_client
        self._sheet_name = sheet_name
        self._group_by = None
        self._sample = None
        self._order_by = None
        self._as_user_id = None
        self._column_aggregations = []
        self._columns = []
        self._tz = force_tz if force_tz else str(get_localzone())
        self._filters = []
        self._limit = 100
        self._view_id = None
        self._no_output = no_output

    def widget(self,
               dashboard_name: str,
               widget_name: str):
        dashboards = self._k.entities.dashboards()
        dashboard = dashboards.get_entity(entity_name=dashboard_name)
        if dashboard is None:
            raise Exception('Dashboard with name {} not found in workspace'.format(dashboard_name))

        widgets = dashboard.get('widgets', [])
        for candidate_widget in widgets:
            candidate_widget_name = candidate_widget.get('displayInformation').get('displayName')
            if candidate_widget_name == widget_name:
                widget_definition = candidate_widget.get('definition')
                layout_id = widget_definition.get('layoutId')
                return self.view_id(layout_id)

        raise Exception('No widget with name {} was found in dashboard {}'.format(widget_name, dashboard_name))

    def view_id(self, view_id: str):
        self._view_id = str(view_id)
        return self

    def as_user_id(self, as_user_id: str):
        self._as_user_id = str(as_user_id)
        return self

    def filter(self, column_filter: KawaFilter):
        self._filters.append(column_filter.to_dict())
        return self

    def group_by(self, *column_names):
        if not self._group_by:
            self._group_by = []

        self._group_by += list(column_names)
        return self

    def order_by(self, column_name: str, ascending: bool = True):
        self._order_by = {
            'columnName': column_name,
            'ascending': ascending
        }
        return self

    def select(self, *columns_or_column_names):
        columns = []
        for column_or_column_name in columns_or_column_names:
            if isinstance(column_or_column_name, KawaColumn):
                columns.append(column_or_column_name.to_dict())
            elif isinstance(column_or_column_name, TreeNode):
                columns.append(column_or_column_name)
            else:
                column = KawaColumn(column_name=str(column_or_column_name))
                columns.append(column.to_dict())

        self._columns = columns
        return self

    def limit(self, limit: int):
        self._limit = limit
        return self

    def no_limit(self):
        self._limit = -1
        return self

    def agg(self, *column_aggregations):
        self._column_aggregations = [c.to_dict() for c in column_aggregations]
        return self

    def sample(self,
               sampler: str,
               how_many_buckets: int = 10,
               bucket_size: int = 10,
               buckets: typing.Optional[list[int]] = None,
               column_name: typing.Optional[str] = None):
        self._sample = {
            'columnName': column_name,
            'sampler': sampler,
            'howManyBuckets': how_many_buckets,
            'bucketSize': bucket_size,
            'buckets': buckets
        }
        return self

    def as_chart(self, chart_name='Generated chart', standalone=True):

        dsl_as_dict = self._transform_to_dict()

        view = self._k.commands.run_command(
            command_name='CreateLayoutFromDsl',
            command_parameters={
                'dsl': dsl_as_dict,
                'chartName': chart_name,
            }
        )

        return view

    def compute(self,
                skip_cache: bool = False,
                use_group_names: bool = False):

        sheet = self._load_sheet()
        self._columns = self.convert_tree_nodes_to_computed_columns(self._columns, sheet)

        return self.collect(
            use_group_names=use_group_names,
            skip_cache=skip_cache)

    @staticmethod
    def convert_tree_nodes_to_computed_columns(columns: list, sheet):
        result = []
        for i, column in enumerate(columns):
            if isinstance(column, TreeNode):
                column_name = column.name if column.name is not None else f"col{i}"
                xml_syntactic_tree = column.to_xml(sheet, parameters=[])
                kawa_column = KawaColumn(column_name=column_name, xml_syntactic_tree=xml_syntactic_tree)
                result.append(kawa_column.to_dict())
            else:
                result.append(column)
        return result

    def collect(self,
                skip_cache: bool = False,
                use_group_names: bool = False):

        url = '{}/computation/compute-from-dsl'.format(self._k.kawa_api_url)

        data = self._transform_to_dict(skip_cache)
        response = self._k.post(url=url, data=data, stream=True)

        if self._no_output:
            return

        with pyarrow.ipc.open_stream(response.content) as reader:
            df = reader.read_pandas()

        if use_group_names and self._group_by:
            for group_id, group_name in enumerate(self._group_by):
                prev_name = f'grouping({group_id})'
                new_name = f'group({group_id}) {group_name}'
                df.rename(columns={prev_name: new_name}, inplace=True)

        return df

    def profile(self, column_name: typing.Optional[str] = None):
        sheet = self._load_sheet()
        all_columns = sheet.get('indicatorColumns', []) + sheet.get('computedColumns', [])
        data = {
            'sheetId': sheet.get('id'),
        }

        if column_name:
            for c in all_columns:
                if c['displayInformation']['displayName'] == column_name:
                    data['columnId'] = c['columnId']

        sheet_profile = self._k.post(
            url=f'{self._k.kawa_api_url}/computation/compute-sheet-stats',
            data=data,
        )

        profile_with_column_names = {}
        for column_id, column_profile in sheet_profile.items():
            matching_column_names = [
                c['displayInformation']['displayName']
                for c in all_columns
                if c['columnId'] == column_id]

            if matching_column_names:
                profile_with_column_names[matching_column_names[0]] = column_profile

        return profile_with_column_names

    def schema(self):
        sheet = self._load_sheet()
        indicator_columns = sheet.get('indicatorColumns', [])
        computed_columns = sheet.get('computedColumns', [])

        return [{c['displayInformation']['displayName']: c['type']} for c in
                [*indicator_columns, *computed_columns]]

    def _transform_to_dict(self, skip_cache=False, ):

        body = {
            'sheetName': self._sheet_name,
            'timeZone': self._tz,
            'limit': self._limit,
            'skipCache': skip_cache,
            'noOutput': self._no_output
        }

        if self._group_by is not None:
            body['groupBy'] = self._group_by

        if self._sample is not None:
            body['sample'] = self._sample

        if self._column_aggregations:
            body['aggregation'] = {'columns': self._column_aggregations}

        if self._columns:
            body['select'] = {'columns': self._columns}

        if self._filters:
            body['filters'] = self._filters

        if self._order_by:
            body['orderBy'] = self._order_by

        if self._view_id:
            body['viewId'] = self._view_id

        if self._as_user_id:
            body['asUserId'] = self._as_user_id

        return body

    def _load_sheet(self):
        sheet = self._k.entities.sheets().get_entity(self._sheet_name)
        if not sheet:
            raise Exception('Sheet with name {} was not found in the current workspace'.format(self._sheet_name))
        return sheet



class KawaConfiguration:

    def __init__(self, kawa_client):
        self._k = kawa_client
        self._configuration_resource_base_url = 'backoffice/configuration'

    def _load_configuration(self, configuration_type):
        try:
            return self._k.get(
                '{}/{}/{}'.format(self._k.kawa_api_url, self._configuration_resource_base_url, configuration_type))
        except:
            return None

    def load_flight_configuration(self):
        return self._load_configuration('ArrowFlightConfiguration')

    def replace_flight_configuration(self, host, port, tls=False, polling_interval=30):
        return self._k.commands.replace_configuration(
            configuration_type='ArrowFlightConfiguration',
            payload={
                'host': host,
                'port': int(port),
                'tls': bool(tls),
                'pollingIntervalSeconds': int(polling_interval)
            }
        )



from enum import Enum

import tzlocal

from .errors import ConflictError
import json
from datetime import date


class SourceControlType(Enum):
    GIT_LAB = 'GIT_LAB'
    GIT_HUB = 'GIT_HUB'


class KawaCommands:

    def __init__(self, kawa_client):
        self._k = kawa_client
        self._secured_commands_url_fragment = '/commands/secured/run'

    def _run_command(self, command_name, command_parameters):
        url = '{}/{}'.format(self._k.kawa_api_url, self._secured_commands_url_fragment)
        r = self._k.post(url, {
            'command': command_name,
            'parameters': command_parameters
        })
        return r

    def _run_idempotent_command(self, command_name, command_parameters, idempotency_key, entity_kind):
        try:
            return self._run_command(command_name, command_parameters)
        except ConflictError:
            # Idempotency: in case of conflict, fetch the entity
            return self._k.entities \
                .entities_of_kind(entity_kind) \
                .get_entity(idempotency_key)

    @staticmethod
    def _entity_id(entity_or_entity_id):
        return entity_or_entity_id if isinstance(entity_or_entity_id, str) else entity_or_entity_id.get('id')

    def run_command(self, command_name: str, command_parameters):
        return self._run_command(command_name, command_parameters)

    """ Misc commands """

    def replace_communication_provider_type(self, configuration_provider_type='LOG'):
        """
        :param configuration_provider_type: accepted values are:
            LOG: No message is sent, just printed out in the server logs (default)
            SMTP: Send emails via SMTP
            MAILGUN: Send email vio the Mailgun REST API
        :return:
        """
        return self._run_command(command_name='ReplaceCommunicationProviderType',
                                 command_parameters={
                                     'communicationProviderType': configuration_provider_type,
                                 })

    def replace_configuration(self, configuration_type, payload):
        return self._run_command(command_name='ReplaceConfiguration',
                                 command_parameters={
                                     'configurationClassSimpleName': configuration_type,
                                     'configurationContent': payload
                                 })

    def toggle_feature(self, feature_name, is_enabled=True):
        return self._run_command(command_name='ToggleFeature',
                                 command_parameters={
                                     'featureName': feature_name,
                                     'enabled': is_enabled
                                 })

    def add_script_runner(self, name, host, port, private_key=None, tls=False):
        return self._run_idempotent_command(command_name='CreateScriptRunner',
                                            idempotency_key=name,
                                            entity_kind='script-runners',
                                            command_parameters={
                                                'name': name,
                                                'host': host,
                                                'port': port,
                                                'tls': tls,
                                                'clearTextPrivateKey': private_key
                                            })

    def toggle_script_runner(self, enabled=True):
        runners = self._k.entities.script_runners().list_entities()
        if not runners:
            raise Exception('No runner found in the workspace')
        return self._run_command(command_name='ToggleScriptRunner',
                                 command_parameters={
                                     'scriptRunnerEntityId': runners[0].get('id'),
                                     'enabled': enabled
                                 })

    def replace_script_runner_location(self, host='localhost', port=8815, tls=True):
        runners = self._k.entities.script_runners().list_entities()
        if not runners:
            raise Exception('No runner found in the workspace')
        return self._run_command(command_name='ReplaceScriptRunnerLocation',
                                 command_parameters={
                                     'scriptRunnerEntityId': runners[0].get('id'),
                                     'host': host,
                                     'port': port,
                                     'tls': tls,
                                 })

    def replace_script_runner_private_key(self, private_key):
        runners = self._k.entities.script_runners().list_entities()
        if not runners:
            raise Exception('No runner found in the workspace')
        return self._run_command(command_name='ReplaceScriptRunnerPrivateKey',
                                 command_parameters={
                                     'scriptRunnerEntityId': runners[0].get('id'),
                                     'clearTextPrivateKey': private_key,
                                 })

    """User and permission management"""

    def get_principal(self, email_or_unique_id: str):
        principal = self._k.entities.principals().find_first_entity_by_attribute('uniqueId', email_or_unique_id)
        if not principal:
            principal = self._k.entities.principals().find_first_entity_by_attribute('email', email_or_unique_id)
        if not principal:
            raise ValueError(f'Principal with email or unique ID {email_or_unique_id} was not found')

        return principal

    def get_principal_entity_id(self, email_or_unique_id: str) -> str:
        principal = self.get_principal(email_or_unique_id)
        return self._entity_id(principal)

    def replace_user_role(self, email_or_unique_id: str, new_role: str):
        return self._run_command(command_name='ReplaceUserRole',
                                 command_parameters={
                                     'principalId': self.get_principal_entity_id(email_or_unique_id),
                                     'role': new_role
                                 })

    def change_user_status(self, email_or_unique_id: str, active: bool = True):
        return self._run_command(command_name='ReplaceUserStatus',
                                 command_parameters={
                                     'principalId': self.get_principal_entity_id(email_or_unique_id),
                                     'active': active
                                 })

    def toggle_users_status(self, user_email_or_id_list: list[str], active: bool = False):
        for email_or_unique_id in user_email_or_id_list:
            self._run_command(command_name='ReplaceUserStatus',
                              command_parameters={
                                  'principalId': self.get_principal_entity_id(email_or_unique_id),
                                  'active': active
                              })

    def replace_access_policy_for_datasource(self, datasource, email_or_unique_id: str, access_policy):
        return self._run_command(command_name='ReplaceAccessPolicyForDataSource',
                                 command_parameters={
                                     "principalId": self.get_principal_entity_id(email_or_unique_id),
                                     "dataSourceId": self._entity_id(datasource),
                                     "accessPolicy": access_policy
                                 })

    def assign_user_to_perimeters(self, email_or_unique_id: str, perimeter_names):
        return self._run_command(command_name='AssignPrincipalToPerimeters',
                                 command_parameters={
                                     'principalEntityId': self.get_principal_entity_id(email_or_unique_id),
                                     'memberOfPerimeterNames': perimeter_names or []
                                 })

    def create_users(self, user_id_list: list[str]):
        return self._run_command(command_name='CreateActivePrincipals',
                                 command_parameters={
                                     'principalIds': user_id_list
                                 })

    def change_user_password(self, email_or_unique_id: str, new_password: str):
        return self._run_command(command_name='AdminChangeUserPassword',
                                 command_parameters={
                                     'principalId': self.get_principal_entity_id(email_or_unique_id),
                                     'newPassword': new_password,
                                 })

    def create_user_with_email_and_password(self, email, password, first_name=None, last_name=None):
        user_tech_id = self._k.entities.principals().find_entities_by_attribute('uniqueId', [email])
        if user_tech_id:
            return
        return self._run_command(command_name='CreateActivePrincipalWithPassword',
                                 command_parameters={
                                     'email': email.lower(),
                                     'password': password,
                                     'firstName': first_name or email,
                                     'lastName': last_name or email
                                 })

    def replace_forbidden_data_types(self, user_email_or_id_list: list[str], forbidden_types):
        for email_or_unique_id in user_email_or_id_list:
            self._run_command(command_name='ReplaceForbiddenDataTypes',
                              command_parameters={
                                  'principalId': self.get_principal_entity_id(email_or_unique_id),
                                  'forbiddenTypes': forbidden_types
                              })

    def replace_permissions(self, user_email_or_id_list: list[str], permissions):
        for email_or_unique_id in user_email_or_id_list:
            self._run_command(command_name='ReplaceApplicationWidePermissions',
                              command_parameters={
                                  'principalId': self.get_principal_entity_id(email_or_unique_id),
                                  'permissions': permissions
                              })

    def toggle_workspace_visibility(self, workspace_id, public=False):
        return self._run_command(command_name='ToggleWorkspaceVisibility',
                                 command_parameters={
                                     'workspaceId': workspace_id,
                                     'isPublic': public
                                 })

    def add_user_to_white_list(self, user_id):
        return self._run_command(command_name='AddUserToWhiteList',
                                 command_parameters={
                                     'userId': user_id.lower(),
                                 })

    def remove_user_from_white_list(self, user_id):
        return self._run_command(command_name='RemoveUserFromWhiteList',
                                 command_parameters={
                                     'userId': user_id.lower(),
                                 })

    def create_users_in_workspace(self,
                                  workspace_name,
                                  user_list,
                                  add_users_to_whitelist=False):
        """
        Create users, create the workspace if it does not exist and put the users in it.
        :param workspace_name: Workspace name to create/load
        :param add_users_to_whitelist: If set to True, will proceed to update the whitelist
        :param user_list: a list of users:
        [{
          'email': 'lucius.fox@wayne-enterprises.com',
          'password': 'LuciusF0x123',
          'first_name': 'Lucius',
          'last_name': 'Fox',
        }]
        """
        workspace = self.create_workspace(workspace_name)
        for user in user_list:
            self.create_user_with_email_and_password(
                email=user['email'],
                password=user['password'],
                first_name=user['first_name'],
                last_name=user['last_name'],
            )
            if add_users_to_whitelist:
                self.add_user_to_white_list(user['email'])

        self.add_users_to_workspace([user['email'].lower() for user in user_list], workspace['id'], True)

    def create_workspace(self, workspace_name):
        existing_entity = self._k.entities.workspaces().get_entity(workspace_name)

        if existing_entity:
            return existing_entity

        return self._run_command(command_name='CreateWorkspace',
                                 command_parameters={
                                     'displayInformation': {
                                         'displayName': workspace_name,
                                         'description': ''
                                     },
                                 })

    def add_users_to_workspace(self, user_id_list, workspace_id, permissions=None, set_as_current_workspace=True):
        principals = self._k.entities.principals().find_entities_by_attribute('uniqueId', user_id_list)
        principal_entity_ids = [self._entity_id(principal) for principal in principals]
        members = [{'principalId': principal_entity_id, 'permissions': permissions or []} for principal_entity_id in
                   principal_entity_ids]

        self._run_command(command_name='AddMembersToWorkspace',
                          command_parameters={
                              'workspaceId': self._entity_id(workspace_id),
                              'members': members
                          })

        if set_as_current_workspace:
            self._run_command(command_name='SetCurrentWorkspaceForPrincipals',
                              command_parameters={
                                  'workspaceId': self._entity_id(workspace_id),
                                  'principalIds': principal_entity_ids
                              })

    def remove_users_from_workspace(self, user_id_list, workspace_id):
        user_tech_id = self._k.entities.principals().find_entities_by_attribute('uniqueId', user_id_list)
        self._run_command(command_name='RemoveWorkspaceMembers',
                          command_parameters={
                              'workspaceId': str(workspace_id),
                              'principalIds': [self._entity_id(user) for user in user_tech_id]
                          })

    """DataSources"""

    def delete_datasource(self, datasource):
        self._run_command(command_name='ArchiveDataSourceAndDeleteAssociatedData',
                          command_parameters={'dataSourceId': self._entity_id(datasource)})

    def create_datasource(self, datasource):
        return self._run_idempotent_command(command_name='CreatePythonDataSource',
                                            idempotency_key=datasource['displayInformation']['displayName'],
                                            entity_kind='datasources',
                                            command_parameters={'dataSource': datasource})

    def create_private_join_datasource(self, datasource):
        return self._run_idempotent_command(command_name='CreatePythonPrivateJoinDataSource',
                                            idempotency_key=datasource['displayInformation']['displayName'],
                                            # TODO THIERRY ########## python-columns: wrong?
                                            entity_kind='datasources',
                                            command_parameters={'dataSource': datasource})

    def add_indicators_to_datasource(self, datasource, new_indicators):
        return self._run_command(command_name='AddIndicatorsToDataSource',
                                 command_parameters={
                                     'dataSourceId': self._entity_id(datasource),
                                     'newIndicators': new_indicators
                                 })

    def share_datasource(self, datasource):
        return self._run_command(command_name='UpdateDataSourceShareStatus',
                                 command_parameters={
                                     'dataSourceId': self._entity_id(datasource),
                                     'shared': True
                                 })

    def drop_date_partition(self, datasource, date_partition: date):
        return self._run_command(command_name='DropDatePartition',
                                 command_parameters={
                                     'dataSourceId': self._entity_id(datasource),
                                     'datePartition': (date_partition - date(1970, 1, 1)).days
                                 })

    def replace_datasource_primary_keys(self, datasource, new_primary_keys, partition_key=None, partition_sampler=None):
        pk = [{'indicatorId': indicator_id} for indicator_id in new_primary_keys]
        datasource_id = self._entity_id(datasource)
        datasource = self._k.entities.datasources().get_entity(datasource_id)
        warehouse_storage_config = datasource.get('storageConfiguration', {}).get('warehouseStorageConfiguration')

        if warehouse_storage_config:
            current_keys = warehouse_storage_config.get('sortingKeys', [])
            current_partition_key = warehouse_storage_config.get('partitionIndicator')
            current_sampler = warehouse_storage_config.get('partitionSampler', {}).get('dateSamplerType')

            if (current_keys == new_primary_keys
                    and partition_key == current_partition_key
                    and current_sampler == partition_sampler):
                # Idempotency: Configuration did not change
                return datasource

        return self._run_command(command_name='ReplaceDataSourcePrimaryKeys',
                                 command_parameters={
                                     'dataSourceId': self._entity_id(datasource),
                                     'newPrimaryKeys': pk,
                                     'partitionIndicator': partition_key,
                                     'partitionSampler': partition_sampler
                                 })

    def delete_data(self, datasource, delete_where):
        filters = [where.to_dict() for where in delete_where]
        return self._run_command(command_name='DeleteData',
                                 command_parameters={
                                     'dataSourceId': self._entity_id(datasource),
                                     'deleteDataFilters': filters
                                 })

    def force_entries_deduplication(self, datasource):
        return self._run_command(command_name='ForceEntriesDeduplication',
                                 command_parameters={
                                     'dataSourceId': self._entity_id(datasource)
                                 })

    """Data providers"""

    def replace_data_provider_connection_security_parameters(self, data_provider, ssl, kerberos):
        return self._run_command(command_name='ReplaceDataProviderConnectionSecurityParameters',
                                 command_parameters={
                                     'dataProviderId': self._entity_id(data_provider),
                                     'connectionSecurityParameters': {
                                         'useSsl': ssl,
                                         'useKerberos': kerberos
                                     }
                                 })

    def replace_data_provider_connection_parameters(self, data_provider, connection_parameters):
        return self._run_command(command_name='ReplaceDataProviderConnectionParameters',
                                 command_parameters={
                                     'dataProviderId': self._entity_id(data_provider),
                                     'connectionParameters': connection_parameters
                                 })

    def archive_data_provider(self, data_provider):
        return self._run_command(command_name='DeleteDataProvider',
                                 command_parameters={
                                     'dataProviderId': self._entity_id(data_provider)
                                 })

    def create_data_provider(self,
                             provider_name,
                             connection_parameters,
                             provider_type='JDBC',
                             restricted=False,
                             ssl=False,
                             kerberos=False):
        return self._run_idempotent_command(command_name='CreateDataProvider',
                                            idempotency_key=provider_name,
                                            entity_kind='data-providers',
                                            command_parameters={
                                                'displayInformation': {
                                                    'displayName': provider_name,
                                                    'description': ''
                                                },
                                                'connectionSecurityParameters': {
                                                    'useSsl': ssl,
                                                    'useKerberos': kerberos
                                                },
                                                'restricted': restricted,
                                                'dataProviderType': provider_type,
                                                'connectionParameters': connection_parameters
                                            })

    def create_or_update_data_provider(self,
                                       provider_name,
                                       connection_parameters,
                                       provider_type='JDBC',
                                       check_health=True,
                                       restricted=False,
                                       ssl=False,
                                       kerberos=False):

        provider = self.create_data_provider(provider_name, connection_parameters, provider_type, restricted, ssl,
                                             kerberos)

        self.replace_data_provider_connection_security_parameters(provider, ssl, kerberos)
        self.replace_data_provider_connection_parameters(provider, connection_parameters)

        if check_health:
            print('Testing connection...')
            health = self._k.data_providers.check_provider_health(provider)
            print(health)

    """Data and Connections"""

    def load_data(self, datasource_id, rows, is_incremental, auto_increment_offset=0, loading_session_id=None):
        return self._run_command(command_name='LoadExternalData',
                                 command_parameters={
                                     'loadingMode': 'INCREMENTAL' if is_incremental else 'RESET_BEFORE_INSERT',
                                     'loadingAdapterName': 'CLICKHOUSE',
                                     'rows': rows,
                                     'loadingSessionId': loading_session_id,
                                     'autoIncrementOffset': auto_increment_offset,
                                     'dataSourceId': self._entity_id(datasource_id)
                                 })

    """Layouts, Sheets and Attributes"""

    def create_attribute(self, attribute_name):
        return self._run_idempotent_command(command_name='CreateAttribute',
                                            idempotency_key=attribute_name,
                                            entity_kind='attributes',
                                            command_parameters={
                                                'attribute': {
                                                    'displayInformation': {
                                                        'displayName': attribute_name
                                                    }
                                                }
                                            })

    def delete_sheet(self, sheet):
        self._run_command(command_name='DeleteSheet',
                          command_parameters={
                              'sheetId': self._entity_id(sheet),
                          })

    def create_sheet(self, sheet_name, datasource):
        created_sheet = self._run_command(command_name='CreateSimpleSheet',
                                          command_parameters={
                                              'shared': False,
                                              'createDefaultLayout': True,
                                              'displayInformation': {
                                                  'displayName': sheet_name,
                                                  'description': ''
                                              },
                                              'datasourceId': [{
                                                  'targetDataSourceId': self._entity_id(datasource),
                                                  'foreignKeyNames': [],
                                                  'defaultValueForAttributes': []
                                              }]})

        print('Sheet {} was created: {}/workspaces/{}/sheets/{}/views/{}'.format(sheet_name,
                                                                                 self._k.kawa_api_url,
                                                                                 self._k.active_workspace_id,
                                                                                 created_sheet.get('id'),
                                                                                 created_sheet.get('layoutIds', [])[0]))

        return created_sheet

    def delete_view(self, view):
        self._run_command(command_name='DeleteLayout',
                          command_parameters={'layoutId': self._entity_id(view)})

    def share_sheet_and_layouts(self, sheet_id: str, layout_id_list: list, teams_ids: list = None):
        """
                Change the share status flag to true for the given sheet and the given list of layouts
                in the workspace
                :param sheet_id: id of the sheet to share
                :param layout_id_list: list of sheet layout to share
                :param teams_ids: list of teams to share with (if not present it will be shared with every member
                of the workspace)
                """
        if not isinstance(teams_ids, list):
            teams_ids = []
        self._run_command(command_name='UpdateSheetAndLayoutsShareStatus',
                          command_parameters={
                              'sheetId': sheet_id,
                              'shared': True,
                              'layoutsShareStatuses': {layout_id: True for layout_id in layout_id_list},
                              'sharedTeamIds': teams_ids
                          })

    def save_layout(self, layout_id: str):
        self._run_command(command_name='SaveLayout',
                          command_parameters={
                              'layoutId': layout_id,
                          })

    def replace_layout_locked_status(self, layout_id: str, locked: bool):
        self._run_command(command_name='ReplaceLayoutLockedStatus',
                          command_parameters={
                              'layoutId': layout_id,
                              'locked': locked
                          })

    """Scripts"""

    def create_script(self, name, tool_kit_name, tool_name):
        return self._run_command(command_name='CreateScript',
                                 command_parameters={
                                     'name': name,
                                     'sourceControlToolConfiguration': {
                                         'toolKitName': tool_kit_name,
                                         'toolName': tool_name
                                     }
                                 })

    def replace_script(self, script_id, name, tool_kit_name, tool_name):
        return self._run_command(command_name='ReplaceScript',
                                 command_parameters={
                                     'scriptId': script_id,
                                     'name': name,
                                     'sourceControlToolConfiguration': {
                                         'toolKitName': tool_kit_name,
                                         'toolName': tool_name
                                     }
                                 })

    def create_python_private_join(self, name, sheet_id, origin_layout_id, script_id, param_mapping):
        return self._run_command(command_name='CreatePythonPrivateJoin',
                                 command_parameters={
                                     'name': name,
                                     'sheetId': sheet_id,
                                     'originLayoutId': origin_layout_id,
                                     'scriptId': script_id,
                                     'paramMapping': param_mapping
                                 })

    def start_python_private_join(self, python_private_join_entity_id, zone_id=None):
        if not zone_id:
            zone_id = tzlocal.get_localzone()
        return self._run_command(command_name='StartPythonPrivateJoin',
                                 command_parameters={
                                     'pythonPrivateJoinId': python_private_join_entity_id,
                                     'zoneId': str(zone_id)
                                 })

    def replace_python_private_join_mapping_and_prefix(self, python_private_join_entity_id, param_mapping, prefix_name):
        return self._run_command(command_name='ReplacePythonPrivateJoinMappingAndPrefix',
                                 command_parameters={
                                     'pythonPrivateJoinId': python_private_join_entity_id,
                                     'paramMapping': param_mapping,
                                     'name': prefix_name
                                 })

    def create_secret_if_not_exist(self, key, value, description):
        return self._run_idempotent_command(command_name='CreateSecret',
                                            command_parameters={
                                                'key': key,
                                                'value': value,
                                                'description': description
                                            },
                                            entity_kind='secrets',
                                            idempotency_key=key)

    """SourceControl"""

    def set_source_control_config_to_workspace(self,
                                               workspace_id: str,
                                               api_end_point: str,
                                               token: str,
                                               repo_name: str,
                                               branch_name: str,
                                               source_control_type: SourceControlType):
        return self._run_command(command_name='SetSourceControlApiConfigurationToWorkspace',
                                 command_parameters={
                                     'workspaceId': workspace_id,
                                     'apiEndpoint': api_end_point,
                                     'token': token,
                                     'repoName': repo_name,
                                     'branchName': branch_name,
                                     'sourceControlType': source_control_type.value,
                                 })

    """License"""

    def set_license(self, path_to_license_file):
        with open(path_to_license_file, 'r') as f:
            licence = json.loads(f.read())
            self.run_command(
                command_name='SetLicense',
                command_parameters=licence
            )



import pandas as pd
from datetime import date
import uuid

from ..kawa_base_e2e_test import KawaBaseTest


class TestPartitionManagement(KawaBaseTest):

    def test_that_primary_keys_order_is_respected_on_datasource_creation(self):
        # given
        unique_id = 'resource_{}'.format(uuid.uuid4())
        dtf1 = pd.DataFrame([{
            'int1': 1,
            'date': date(2024, 1, 1),
            'int3': 1,
            'int2': 1,
            'int4': 1}])

        loader = self.kawa.new_data_loader(
            df=dtf1,
            datasource_name=unique_id)

        # when
        datasource = loader.create_datasource(primary_keys=['int1', 'date', 'int3', 'int2', 'int4'])
        create_table_before_indexing = self.kawa.entities.get_datasource_schema(
            datasource_id=datasource.get('id')).get('schema')

        # then
        self.assertIn('PRIMARY KEY (int1, date, int3, int2, int4)', create_table_before_indexing)
        self.assertIn('ORDER BY (int1, date, int3, int2, int4)', create_table_before_indexing)

    def test_that_primary_keys_order_is_respected_when_updating_them(self):
        # given
        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader = self.kawa.new_data_loader(
            df=pd.DataFrame([{
                'int1': 1,
                'date': date(2024, 1, 1),
                'int3': 1,
                'int2': 1,
                'int4': 1}]),
            datasource_name=unique_id)

        # when
        datasource = loader.create_datasource(primary_keys=['int1', 'date', 'int2'])
        create_table_before_indexing = self.kawa.entities.get_datasource_schema(
            datasource_id=datasource.get('id')).get('schema')

        # and
        self.commands.replace_datasource_primary_keys(
            datasource=datasource,
            new_primary_keys=['int2', 'int1', 'int4', 'date', 'int3'])
        create_table_after_indexing = self.kawa.entities.get_datasource_schema(
            datasource_id=datasource.get('id')).get('schema')

        # then
        self.assertIn('PRIMARY KEY (int1, date, int2)', create_table_before_indexing)
        self.assertIn('ORDER BY (int1, date, int2)', create_table_before_indexing)
        self.assertIn('PRIMARY KEY (int2, int1, int4, date, int3)', create_table_after_indexing)
        self.assertIn('ORDER BY (int2, int1, int4, date, int3)', create_table_after_indexing)

    def test_that_a_partition_is_correctly_dropped(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        dtf1 = pd.DataFrame([
            {'date': date(2024, 1, 1), 'int': 1},
            {'date': date(2024, 1, 1), 'int': 2},
            {'date': date(2024, 1, 1), 'int': 3},
            {'date': date(2024, 1, 2), 'int': 4},
            {'date': date(2024, 1, 2), 'int': 5},
            {'date': date(2024, 1, 3), 'int': 6},
            {'date': date(2024, 1, 3), 'int': 7},
        ])

        loader = self.kawa.new_data_loader(df=dtf1, datasource_name=unique_id)
        loader.create_datasource()
        datasource = loader.load_data(reset_before_insert=True, create_sheet=True)
        self.commands.replace_datasource_primary_keys(datasource=datasource,
                                                      new_primary_keys=['date', 'record_id'],
                                                      partition_key='date')

        create_table = self.kawa.entities.get_datasource_schema(datasource_id=datasource.get('id')).get('schema')

        # given
        computed_df_before_partition_drop = self.kawa.sheet(sheet_name=unique_id).select('date', 'int').compute()

        # when
        self.commands.drop_date_partition(datasource=datasource, date_partition=date(2024, 1, 1))
        computed_df_after_partition_drop = self.kawa.sheet(sheet_name=unique_id).select('date', 'int').compute()

        # then: computation before drop is correct
        self.assertListEqual(list(computed_df_before_partition_drop['int']), list(range(1, 8)))
        self.assertListEqual(list(computed_df_before_partition_drop['date']), [
            date(2024, 1, 1),
            date(2024, 1, 1),
            date(2024, 1, 1),
            date(2024, 1, 2),
            date(2024, 1, 2),
            date(2024, 1, 3),
            date(2024, 1, 3)
        ])

        # and: computation after drop is correct
        self.assertListEqual(list(computed_df_after_partition_drop['int']), [4, 5, 6, 7])
        self.assertListEqual(list(computed_df_after_partition_drop['date']), [
            date(2024, 1, 2),
            date(2024, 1, 2),
            date(2024, 1, 3),
            date(2024, 1, 3)
        ])

        # and: check the schema
        self.assertIn('PRIMARY KEY (date, record_id)', create_table)
        self.assertIn('ORDER BY (date, record_id)', create_table)
        self.assertIn('PARTITION BY date', create_table)

    def test_that_a_partition_is_created_with_up_sampling(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        dtf1 = pd.DataFrame([
            {'date': date(2024, 1, 1), 'int': 1},
            {'date': date(2024, 1, 1), 'int': 2},
            {'date': date(2024, 1, 1), 'int': 3},
            {'date': date(2024, 1, 2), 'int': 4},
            {'date': date(2024, 1, 2), 'int': 5},
            {'date': date(2024, 1, 3), 'int': 6},
            {'date': date(2024, 1, 3), 'int': 7},
        ])

        loader = self.kawa.new_data_loader(df=dtf1, datasource_name=unique_id)
        loader.create_datasource()
        datasource = loader.load_data(reset_before_insert=True, create_sheet=True)

        # when
        self.commands.replace_datasource_primary_keys(datasource=datasource,
                                                      new_primary_keys=['date', 'record_id'],
                                                      partition_key='date',
                                                      partition_sampler='YEAR')

        create_table = self.kawa.entities.get_datasource_schema(datasource_id=datasource.get('id')).get('schema')

        # and: check the schema
        self.assertIn('PRIMARY KEY (date, record_id)', create_table)
        self.assertIn('ORDER BY (date, record_id)', create_table)
        self.assertIn('PARTITION BY toString(toYear(addDays(toDate32(0), date)))', create_table)

    def test_that_a_partition_is_maintained_after_reset_data(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        dtf1 = pd.DataFrame([
            {'date': date(2024, 1, 1), 'int': 1},
            {'date': date(2024, 1, 1), 'int': 2},
            {'date': date(2024, 1, 1), 'int': 3},
            {'date': date(2024, 1, 2), 'int': 4},
            {'date': date(2024, 1, 2), 'int': 5},
            {'date': date(2024, 1, 3), 'int': 6},
            {'date': date(2024, 1, 3), 'int': 7},
        ])

        loader = self.kawa.new_data_loader(df=dtf1, datasource_name=unique_id)
        loader.create_datasource()
        datasource = loader.load_data(reset_before_insert=True, create_sheet=True)
        self.commands.replace_datasource_primary_keys(datasource=datasource,
                                                      new_primary_keys=['date', 'record_id'],
                                                      partition_key='date',
                                                      partition_sampler='YEAR')

        # when
        loader2 = self.kawa.new_data_loader(df=dtf1, datasource_name=unique_id)
        datasource = loader2.load_data(reset_before_insert=True, create_sheet=False)

        create_table = self.kawa.entities.get_datasource_schema(datasource_id=datasource.get('id')).get('schema')

        # and: check the schema
        self.assertIn('PRIMARY KEY (date, record_id)', create_table)
        self.assertIn('ORDER BY (date, record_id)', create_table)
        self.assertIn('PARTITION BY toString(toYear(addDays(toDate32(0), date)))', create_table)

    def test_that_replace_datasource_primary_keys_is_idempotent(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        dtf1 = pd.DataFrame([
            {'date': date(2024, 1, 1), 'int': 1},
            {'date': date(2024, 1, 1), 'int': 2},
            {'date': date(2024, 1, 1), 'int': 3},
            {'date': date(2024, 1, 2), 'int': 4},
            {'date': date(2024, 1, 2), 'int': 5},
            {'date': date(2024, 1, 3), 'int': 6},
            {'date': date(2024, 1, 3), 'int': 7},
        ])

        loader = self.kawa.new_data_loader(df=dtf1, datasource_name=unique_id)
        loader.create_datasource()
        ds0 = loader.load_data(reset_before_insert=True, create_sheet=True)

        # when
        ds1 = self.commands.replace_datasource_primary_keys(datasource=ds0,
                                                            new_primary_keys=['date', 'record_id'],
                                                            partition_key='date')

        ds2 = self.commands.replace_datasource_primary_keys(datasource=ds1,
                                                            new_primary_keys=['date', 'record_id'],
                                                            partition_key='date')

        ds3 = self.commands.replace_datasource_primary_keys(datasource=ds1,
                                                            new_primary_keys=['date', 'record_id'],
                                                            partition_key='date',
                                                            partition_sampler='MONTH')

        ds4 = self.commands.replace_datasource_primary_keys(datasource=ds3,
                                                            new_primary_keys=['date', 'record_id'],
                                                            partition_key='date',
                                                            partition_sampler='MONTH')

        ds5 = self.commands.replace_datasource_primary_keys(datasource=ds4,
                                                            new_primary_keys=['record_id', 'date'],
                                                            partition_key='date',
                                                            partition_sampler='MONTH')

        ds6 = self.commands.replace_datasource_primary_keys(datasource=ds5,
                                                            new_primary_keys=['record_id', 'date'])

        ds7 = self.commands.replace_datasource_primary_keys(datasource=ds6,
                                                            new_primary_keys=['record_id', 'date'])

        ds8 = self.commands.replace_datasource_primary_keys(datasource=ds6,
                                                            new_primary_keys=['date', 'record_id'])

        # then
        self.assertEqual(ds0['version'] + 1, ds1['version'])
        self.assertEqual(ds1['version'], ds2['version'])  # Version did not change there
        self.assertEqual(ds2['version'] + 1, ds3['version'])
        self.assertEqual(ds3['version'], ds4['version'])
        self.assertEqual(ds4['version'] + 1, ds5['version'])
        self.assertEqual(ds5['version'] + 1, ds6['version'])
        self.assertEqual(ds6['version'], ds7['version'])
        self.assertEqual(ds7['version'] + 1, ds8['version'])



import pandas as pd
import uuid
from datetime import datetime, time, timedelta
from ...kawa_base_e2e_test import KawaBaseTest
from ....client.kawa_client import KawaClient as kawa
import zoneinfo


class TestComputationDslWithVariousDataSets(KawaBaseTest):

    def test_date_time_filters(self):
        # setup
        eastern_tz = zoneinfo.ZoneInfo('US/Eastern')
        tiny_timeseries = pd.DataFrame([
            {'id': 13, 'time': datetime(2021, 1, 4, 13, 0, 0, tzinfo=eastern_tz)},
            {'id': 14, 'time': datetime(2021, 1, 4, 14, 0, 0, tzinfo=eastern_tz)},
            {'id': 15, 'time': datetime(2021, 1, 4, 15, 0, 0, tzinfo=eastern_tz)},
            {'id': 16, 'time': datetime(2021, 1, 4, 16, 0, 0, tzinfo=eastern_tz)},
        ])

        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader = self.kawa.new_data_loader(df=tiny_timeseries, datasource_name=unique_id)
        loader.create_datasource(primary_keys=['id'])
        loader.load_data(reset_before_insert=True, create_sheet=True)

        # given
        def compute(tz_id, from_inclusive=None, to_inclusive=None):
            time_filter = (kawa
                           .col('time')
                           .datetime_range(from_inclusive=from_inclusive,
                                           to_inclusive=to_inclusive))
            return list((self.kawa
                         .sheet(sheet_name=unique_id, force_tz=tz_id)
                         .select('id')
                         .filter(time_filter)
                         .compute())['id'])

        # expect
        self.assertListEqual(
            compute(
                tz_id='US/Eastern',
                from_inclusive=datetime(2021, 1, 4, 15, 0, 0, tzinfo=eastern_tz)
            ), [15, 16])
        self.assertListEqual(
            compute(
                tz_id='US/Eastern',
                to_inclusive=datetime(2021, 1, 4, 15, 0, 0, tzinfo=eastern_tz)
            ), [13, 14, 15])
        self.assertListEqual(
            compute(
                tz_id='UTC',
                to_inclusive=datetime(2021, 1, 4, 15, 0, 0, tzinfo=eastern_tz)
            ), [13, 14, 15])

    def test_date_time_filters_with_milliseconds(self):
        # setup
        eastern_tz = zoneinfo.ZoneInfo('US/Eastern')
        tiny_timeseries = pd.DataFrame([
            {'id': 1, 'time': datetime(2021, 1, 1, 0, 0, 0, tzinfo=eastern_tz)},
            {'id': 2, 'time': datetime(2021, 1, 1, 0, 0, 0, tzinfo=eastern_tz) + timedelta(milliseconds=250)},
            {'id': 3, 'time': datetime(2021, 1, 1, 0, 0, 0, tzinfo=eastern_tz) + timedelta(milliseconds=500)},
            {'id': 4, 'time': datetime(2021, 1, 1, 0, 0, 0, tzinfo=eastern_tz) + timedelta(milliseconds=750)},
            {'id': 5, 'time': datetime(2021, 1, 1, 0, 0, 1, tzinfo=eastern_tz)},
        ])

        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader = self.kawa.new_data_loader(df=tiny_timeseries, datasource_name=unique_id)
        loader.create_datasource(primary_keys=['id'])
        loader.load_data(reset_before_insert=True, create_sheet=True)

        # given
        def compute(tz_id, from_inclusive=None, to_inclusive=None):
            time_filter = (kawa
                           .col('time')
                           .datetime_range(from_inclusive=from_inclusive,
                                           to_inclusive=to_inclusive))
            return list((self.kawa
                         .sheet(sheet_name=unique_id, force_tz=tz_id)
                         .select('id')
                         .filter(time_filter)
                         .compute())['id'])

        # expect
        self.assertListEqual(
            compute(
                tz_id='US/Eastern',
                from_inclusive=datetime(2021, 1, 1, 0, 0, 0, tzinfo=eastern_tz) + timedelta(milliseconds=250)
            ), [2, 3, 4, 5])
        self.assertListEqual(
            compute(
                tz_id='US/Eastern',
                to_inclusive=datetime(2021, 1, 1, 0, 0, 0, tzinfo=eastern_tz) + timedelta(milliseconds=500)
            ), [1, 2, 3])

    def test_time_range_filters(self):
        # setup
        eastern_tz = zoneinfo.ZoneInfo('US/Eastern')
        tiny_timeseries = pd.DataFrame([
            {'id': 13, 'time': datetime(2021, 1, 4, 13, 0, 0, tzinfo=eastern_tz)},
            {'id': 14, 'time': datetime(2022, 1, 3, 14, 0, 0, tzinfo=eastern_tz)},
            {'id': 15, 'time': datetime(2023, 1, 4, 15, 0, 0, tzinfo=eastern_tz)},
            {'id': 16, 'time': datetime(2024, 1, 1, 16, 0, 0, tzinfo=eastern_tz)},
        ])

        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader = self.kawa.new_data_loader(df=tiny_timeseries, datasource_name=unique_id)
        loader.create_datasource(primary_keys=['id'])
        loader.load_data(reset_before_insert=True, create_sheet=True)

        # given
        def compute(tz_id,
                    from_inclusive=None,
                    from_exclusive=None,
                    to_inclusive=None,
                    to_exclusive=None):
            time_filter = kawa.col('time').time_range(from_inclusive=from_inclusive,
                                                      from_exclusive=from_exclusive,
                                                      to_inclusive=to_inclusive,
                                                      to_exclusive=to_exclusive)
            return list((self.kawa
                         .sheet(sheet_name=unique_id, force_tz=tz_id)
                         .select('id')
                         .filter(time_filter)
                         .compute())['id'])

        # expect
        # Data contains
        # 13:00, 14:00, 15:00, 16:00 in eastern time
        # Which translates to
        # 18:00, 19:00, 20:00, 21:00 in UTC

        # ----------------
        # US/Eastern computations
        # ----------------
        self.assertListEqual(
            compute(
                tz_id='US/Eastern',
                from_inclusive=time(hour=15, minute=0, second=0)
            ), [15, 16])

        self.assertListEqual(
            compute(
                tz_id='US/Eastern',
                from_exclusive=time(hour=15, minute=0, second=0)
            ), [16])

        self.assertListEqual(
            compute(
                tz_id='US/Eastern',
                to_inclusive=time(hour=15, minute=0, second=0)
            ), [13, 14, 15])

        self.assertListEqual(
            compute(
                tz_id='US/Eastern',
                to_exclusive=time(hour=15, minute=0, second=0)
            ), [13, 14])

        self.assertListEqual(
            compute(
                tz_id='US/Eastern',
                from_exclusive=time(hour=13, minute=0, second=0),
                to_exclusive=time(hour=15, minute=0, second=0),
            ), [14])

        self.assertListEqual(
            compute(
                tz_id='US/Eastern',
                from_inclusive=time(hour=13, minute=0, second=0),
                to_exclusive=time(hour=15, minute=0, second=0),
            ), [13, 14])

        self.assertListEqual(
            compute(
                tz_id='US/Eastern',
                from_exclusive=time(hour=13, minute=0, second=0),
                to_inclusive=time(hour=15, minute=0, second=0),
            ), [14, 15])

        self.assertListEqual(
            compute(
                tz_id='US/Eastern',
                from_inclusive=time(hour=15, minute=0, second=0),
                to_inclusive=time(hour=15, minute=0, second=0),
            ), [15])

        self.assertListEqual(
            compute(
                tz_id='US/Eastern',
                from_exclusive=time(hour=15, minute=0, second=0),
                to_inclusive=time(hour=15, minute=0, second=0),
            ), [])

        # ----------------
        # UTC computations
        # ----------------
        self.assertListEqual(
            compute(
                tz_id='UTC',
                # Equivalent of 15:00 Eastern
                from_inclusive=time(hour=20, minute=0, second=0)
            ), [15, 16])

        self.assertListEqual(
            compute(
                tz_id='UTC',
                # Equivalent of 15:00 Eastern
                from_exclusive=time(hour=20, minute=0, second=0)
            ), [16])

        self.assertListEqual(
            compute(
                tz_id='UTC',
                # Equivalent of 15:00 Eastern
                to_inclusive=time(hour=20, minute=0, second=0)
            ), [13, 14, 15])

        self.assertListEqual(
            compute(
                tz_id='UTC',
                # Equivalent of 15:00 Eastern
                to_exclusive=time(hour=20, minute=0, second=0)
            ), [13, 14])



import typing
from dataclasses import dataclass
import datetime

import xml.etree.ElementTree as ET


@dataclass
class TreeNode:
    id: str
    children: list['TreeNode']
    name: typing.Optional[str] = None
    value: typing.Optional[str | int | float | bool | datetime.datetime | datetime.date] = None
    to_xml_func: any = None
    column_index = {}

    def to_xml(self, sheet, parameters):

        column_index = {}
        parameter_index = {}

        for column in sheet.get('indicatorColumns', []) + sheet.get('computedColumns', []):
            column_index[column.get('displayInformation').get('displayName')] = column.get('columnId')

        for parameter in parameters:
            parameter_index[parameter.get('name')] = parameter.get('id')

        self._inject_column_ids(column_index, parameter_index)
        raw_xml = '<xml>' + self._to_xml() + '</xml>'
        element = ET.XML(raw_xml)
        ET.indent(element)
        return ET.tostring(element, encoding='unicode')

    def alias(self, alias: str):
        self.name = alias
        return self

    def _inject_column_ids(self,
                           column_index: dict[str, str],
                           parameter_index: dict[str, str]):
        for child in self.children:
            # In-place mutate the identifiers
            if child.id == 'column_identifier':
                column_name = child.value.strip()
                column_id = column_index.get(column_name)
                if not column_id:
                    raise ValueError(f'The column with name "{column_name}" was not found')
                child.value = column_id
            elif child.id == 'parameter_identifier':
                parameter_name = child.value.strip()
                parameter_id = parameter_index.get(parameter_name)
                if not parameter_id:
                    raise ValueError(f'The parameter with name "{parameter_name}" was not found')
                child.value = parameter_id
            else:
                child._inject_column_ids(column_index, parameter_index)

    def _to_xml(self):
        node = self
        return self.to_xml_func(node)

    def binary_expression_xml(self):
        return f'''
        <block type="BINARY_EXPRESSION">
            <field name="operation_id">{self.id}</field>
            <value name="left">{self.children[0]._to_xml()}</value>
            <value name="right">{self.children[1]._to_xml()}</value>
        </block>'''

    def call_expression_xml(self):
        inputs = []
        input_counter = 0
        for child in self.children:
            input_counter += 1
            inputs.append(f'<value name="input{input_counter}">{child._to_xml()}</value>')

        inputs_xml = '\n'.join(inputs)
        return f'''
            <block type="CALL_EXPRESSION">
                <field name="operation_id">{self.id}</field>
                {inputs_xml}
            </block>'''

    def literal_xml(self):
        if isinstance(self.value, str):
            kawa_type = 'text'
        elif isinstance(self.value, float):
            kawa_type = 'decimal'
        elif isinstance(self.value, bool):
            kawa_type = 'boolean'
        else:
            kawa_type = 'integer'

        return f'''
            <block type="LITERAL">
                <field name="type">{kawa_type}</field>
                <field name="value">{self.value}</field>
            </block>
        '''

    def identifier_xml(self):
        return f'''
              <block type="IDENTIFIER">
                <field name="type">text</field>
                <field name="reference">{self.value}</field>
            </block>
           '''


def binary_expression(operation_id: str, node1: TreeNode, node2: TreeNode) -> TreeNode:
    return TreeNode(
        id=operation_id,
        children=[node1, node2],
        to_xml_func=lambda node: node.binary_expression_xml(),
    )


def call_expression(operation_id: str, children: list[TreeNode]) -> TreeNode:
    return TreeNode(
        id=operation_id,
        children=children,
        to_xml_func=lambda node: node.call_expression_xml(),
    )


def val(value: str | int | float | bool | datetime.datetime | datetime.date) -> TreeNode:
    if value == ' ':
        return space()
    else:
        return TreeNode(
            id='val',
            children=[],
            value=value,
            to_xml_func=lambda node: node.literal_xml(),
        )


def col_identifier(identifier_value: str) -> TreeNode:
    return TreeNode(
        id='column_identifier',
        children=[],
        value=identifier_value,
        to_xml_func=lambda node: node.identifier_xml(),
    )


def param_identifier(identifier_value: str) -> TreeNode:
    return TreeNode(
        id='parameter_identifier',
        children=[],
        value=identifier_value,
        to_xml_func=lambda node: node.identifier_xml(),
    )


def col(column_name: str): return call_expression('COLUMN', [col_identifier(column_name)])


def param(parameter_name: str): return call_expression('PARAMETER', [param_identifier(parameter_name)])


def length(input_node: TreeNode): return call_expression('LEN', [input_node])


def first(input_node: TreeNode): return call_expression('FIRST', [input_node])


def total(input_node: TreeNode): return call_expression('SUM', [input_node])


def average(input_node: TreeNode): return call_expression('AVG', [input_node])


def str_(input_node: TreeNode): return call_expression('TEXT', [input_node])


def extract(input_text: TreeNode, regexp: TreeNode):
    return call_expression('EXTRACT', [input_text, regexp])


def add(node1: TreeNode, node2: TreeNode): return binary_expression('ADDITION', node1, node2)


def sub(node1: TreeNode, node2: TreeNode): return binary_expression('SUBTRACTION', node1, node2)


def mul(node1: TreeNode, node2: TreeNode): return binary_expression('MULTIPLICATION', node1, node2)


def div(node1: TreeNode, node2: TreeNode): return binary_expression('DIVISION', node1, node2)


def ge(node1: TreeNode, node2: TreeNode): return binary_expression('GREATER_THAN_OR_EQUAL', node1, node2)


def gt(node1: TreeNode, node2: TreeNode): return binary_expression('STRICTLY_GREATER_THAN', node1, node2)


def le(node1: TreeNode, node2: TreeNode): return binary_expression('LESSER_THAN_OR_EQUAL', node1, node2)


def lt(node1: TreeNode, node2: TreeNode): return binary_expression('STRICTLY_LESSER_THAN', node1, node2)


def eq(node1: TreeNode, node2: TreeNode): return binary_expression('EQUAL', node1, node2)


def ne(node1: TreeNode, node2: TreeNode): return binary_expression('NOT_EQUAL', node1, node2)


def space(): return call_expression('SPACE', [])


def power(node1: TreeNode, node2: TreeNode): return call_expression('POWER', [node1, node2])


def starts_with(node1: TreeNode, node2: TreeNode): return call_expression('BEGINS_WITH', [node1, node2])


def ends_with(node1: TreeNode, node2: TreeNode): return call_expression('ENDS_WITH', [node1, node2])


def contains(node1: TreeNode, node2: TreeNode): return call_expression('CONTAINS', [node1, node2])


def left(node1: TreeNode, node2: TreeNode): return call_expression('LEFT', [node1, node2])


def right(node1: TreeNode, node2: TreeNode): return call_expression('RIGHT', [node1, node2])


def find(node1: TreeNode, node2: TreeNode): return call_expression('FIND', [node1, node2])


def substring(*nodes: list[TreeNode]): return call_expression('SUBSTRING', list(nodes))


def replace_all(node1: TreeNode, node2: TreeNode, node3: TreeNode):
    return call_expression('SUBSTITUTE_ALL', [node1, node2, node3])


def replace_first(node1: TreeNode, node2: TreeNode, node3: TreeNode):
    return call_expression('SUBSTITUTE', [node1, node2, node3])


def abs_(node1: TreeNode): return call_expression('ABS', [node1])


def not_(node1: TreeNode): return call_expression('NOT', [node1])


def in_list(*nodes: list[TreeNode]): return call_expression('IN_LIST', list(nodes))


def largest(node1: TreeNode, node2: TreeNode): return call_expression('LARGEST', [node1, node2])


def smallest(node1: TreeNode, node2: TreeNode): return call_expression('SMALLEST', [node1, node2])


def empty(node1: TreeNode): return call_expression('IS_EMPTY', [node1])


def not_empty(node1: TreeNode): return call_expression('IS_NOT_EMPTY', [node1])


def lower(node1: TreeNode): return call_expression('LOWER', [node1])


def upper(node1: TreeNode): return call_expression('UPPER', [node1])


def trim(node1: TreeNode): return call_expression('TRIM', [node1])


def concat(*nodes: list[TreeNode]): return call_expression('CONCAT', list(nodes))


def if_then_else(*nodes: list[TreeNode]): return call_expression('IF', list(nodes))


def or_(*nodes: list[TreeNode]): return call_expression('OR', list(nodes))


def and_(*nodes: list[TreeNode]): return call_expression('AND', list(nodes))



import glob
import os.path
import re
import shutil
import time
import traceback
import uuid
from datetime import date
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.ipc as ipc
import concurrent.futures
import io
import numpy


class KawaDataLoader:

    def __init__(self,
                 kawa_client,
                 df,
                 datasource_name=None,
                 datasource_is_shared=False,
                 datasource_id=None,
                 copy_df=True,
                 arrow_table=None,
                 ):
        if datasource_id is None and datasource_name is None:
            raise Exception('To build a KawaDataLoader you need either a datasource_name or a datasource_id')
        self._k = kawa_client

        if (arrow_table is None) == (df is None):
            raise Exception('To build a KawaDataLoader you need either a pandas dataframe or an arrow table')

        self._use_arrow = arrow_table is not None
        self._arrow_table = arrow_table

        if self._use_arrow:
            self._df = arrow_table.slice(0, 1000).to_pandas()
        else:
            self._df = df.copy() if copy_df else df

        self._shared = datasource_is_shared
        self._name = '' if datasource_name is None else datasource_name.strip()
        self._datasource_id = datasource_id

    def create_datasource(self, primary_keys=None):
        """
        If the datasource with name == datasource_name does not exist in the current
        workspace, create it.
        This command is idempotent.
        Once the datasource exists, it will have no effect at all.

        The indicators of the datasource will be deduced from the columns of the dataframe.

        :param: primary_keys:
        Defines the list of columns that will be considered as primary key.
        If left empty, KAWA will generate a record_id which will act as an auto increment primary key.

        IMPORTANT: Those primary keys will be imported in the same order as they appear in the dataframe.

        :return: The created or existing datasource
        """
        if self._datasource_id:
            raise Exception('Cannot create data source when data source id has been provided')
        existing_datasource = self._k.entities.datasources().get_entity(entity_id_or_name=self._name)
        if existing_datasource:
            # Idempotent
            return existing_datasource
        else:
            ds = self._define_data_source_from_df(primary_keys=primary_keys)
            return self._k.commands.create_datasource(datasource=ds)

    def add_new_indicators_to_datasource(self):
        """
        Adds to the existing datasource (identified by its name) all the columns of the dataframe that
        were not already there.
        This command will only add new indicators to the datasource and will have no other effect.

        :raises: Exception is raised if the datasource does not exist in the current workspace
        :return: The updated datasource
        """
        existing_datasource = self._get_data_source_or_raise()

        existing_indicator_id = []
        existing_primary_keys = []

        for indicator in existing_datasource.get('indicators', []):
            indicator_id = indicator.get('indicatorId')
            existing_indicator_id.append(indicator_id)
            key = indicator.get('key', {})
            key_type = key.get('keyType', '')
            if 'PRIMARY' in key_type:
                existing_primary_keys.append(indicator_id)

        # No new primary key here because we only look for new indicators
        # And BE does not support adding PK via the add_indicators_to_datasource command.
        new_datasource_from_df = self._define_data_source_from_df(existing_primary_keys)

        new_indicators = []
        for new_indicator in new_datasource_from_df.get('indicators', []):
            new_indicator_id = new_indicator.get('indicatorId')
            if new_indicator_id not in existing_indicator_id:
                new_indicators.append(new_indicator)

        if new_indicators:
            print('Adding the following indicators: {}'.format([i.get('indicatorId') for i in new_indicators]))
            return self._k.commands.add_indicators_to_datasource(
                datasource=existing_datasource,
                new_indicators=new_indicators)
        else:
            print('No new indicator to add')
            return existing_datasource

    def load_data(self,
                  reset_before_insert=False,
                  create_sheet=False,
                  optimize_after_insert=False,
                  nb_threads=1,
                  parquet_file_list=None,
                  job_id=None):
        """
        Performs the following operations:
        1) Create the datasource if it does not exist, add any missing indicator to the existing one
        2) Send the dataframe to kawa

        :param: reset_before_insert:
        Set to True if the data has to be reset before the load.
        Set to False if the data has to be appended to existing data.

        :param: create_sheet:
        Set to True if a sheet should be created after the load.
        The URL to the sheet will be printed out.

        :param: nb_threads:
        Will split the data in nb_threads partition and load each one from
        a different thread.

        :return:
        The datasource object in which the data was loaded
        """
        created_data_source = self._load_data(optimize_after_insert=optimize_after_insert,
                                              reset_data=reset_before_insert,
                                              nb_threads=nb_threads,
                                              parquet_file_list=parquet_file_list,
                                              session_id_input=job_id)
        if create_sheet:
            self._k.commands.create_sheet(datasource=created_data_source,
                                          sheet_name=self._name)

        return created_data_source

    def _introspect_df(self):
        column_kawa_types = {}

        for column_name in self._df.columns:
            kawa_type = self._extract_kawa_type(column_name)
            column_kawa_types[column_name] = kawa_type

        return column_kawa_types

    def _define_data_source_from_df(self, primary_keys=None):

        defined_pks = primary_keys if primary_keys else []
        indicators = [self._define_indicator(c, c in defined_pks) for c in self._df.columns]

        # Add the auto increment key if there is no specified key
        key_indicators = [i for i in indicators if 'key' in i]
        if not key_indicators:
            indicators.insert(0, {
                'displayInformation': {
                    'displayName': 'record_id'
                },
                'includedInDefaultLayout': False,
                'indicatorId': 'record_id',
                'storageConfig': {
                    'indexed': True,
                    'automaticUniqueValue': True
                },
                'type': 'integer',
                'key': {
                    'keyType': 'PRIMARY_SHARDING_KEY'
                }
            })

        return {
            'shared': self._shared,
            'displayInformation': {
                'displayName': self._name
            },
            'storageConfiguration': {
                'loadingAdapterName': 'CLICKHOUSE'
            },
            'indicators': indicators,
        }

    def _define_indicator(self, column_name, is_primary_key=False):
        indicator = {
            'displayInformation': {
                'displayName': column_name
            },
            'includedInDefaultLayout': True,
            'indicatorId': column_name,
            'storageConfig': {
                'indexed': is_primary_key
            },
            'type': self._extract_kawa_type(column_name)
        }
        if is_primary_key:
            indicator['key'] = {'keyType': 'PRIMARY_SHARDING_KEY'}

        return indicator

    def _extract_kawa_type(self, column_name):
        column = self._df[column_name]
        column_type_name = str(column.dtype)

        if re.match(r'^datetime64', column_type_name):
            return 'date_time'

        if column_type_name == 'object':
            return self._introspect_values(column_name)

        if column_type_name == 'string':
            return 'text'

        if column_type_name == 'bool':
            return 'boolean'

        if re.match(r'^u?int[0-9]*$', column_type_name):
            return 'integer'

        if re.match(r'^float[0-9]*$', column_type_name):
            return 'decimal'

        raise Exception('Column {} with type {} is not supported'.format(column_name, column_type_name))

    def _introspect_values(self, column_name):
        for val in self._df[column_name]:
            if type(val) is str:
                return 'text'
            if type(val) is date:
                return 'date'
            if type(val) is list and val:
                for list_item in val:
                    if type(list_item) is str:
                        return 'list(integer,text)'
                    if type(list_item) is int:
                        return 'list(integer,integer)'
                    if type(list_item) is float:
                        return 'list(integer,decimal)'

        return 'any'

    def _get_data_source_or_raise(self):
        if self._datasource_id:
            datasource = self._k.entities.datasources().get_entity_by_id(self._datasource_id)
            if not datasource:
                raise Exception(
                    'No datasource with id: {} was found in the current workspace'.format(self._datasource_id))
        else:
            datasource = self._k.entities.datasources().get_entity(entity_id_or_name=self._name)
            if not datasource:
                raise Exception('No datasource with name: {} was found in the current workspace'.format(self._name))
        return datasource

    def _load_data(self,
                   show_progress=True,
                   reset_data=True,
                   optimize_after_insert=False,
                   nb_threads=1,
                   parquet_file_list=None,
                   session_id_input=None):

        df = self._df

        datasource = self._get_data_source_or_raise()
        datasource_id = datasource.get('id')

        nb_rows_to_be_inserted = df.shape[0]
        indicators = datasource.get('indicators')
        session_id = session_id_input if session_id_input is not None else str(uuid.uuid4())
        print('Starting an ingestion session with id={}'.format(session_id))

        # URLs for ingestion session
        data_format = 'ArrowStream' if self._use_arrow else 'parquet'
        query_params = 'datasource={}&format={}&reset={}&session={}&optimize={}'.format(datasource_id,
                                                                                        data_format,
                                                                                        reset_data,
                                                                                        session_id,
                                                                                        optimize_after_insert)
        prepare_url = '{}/ingestion/prepare?{}'.format(self._k.kawa_api_url, query_params)
        ingest_url = '{}/ingestion/upload?{}'.format(self._k.kawa_api_url, query_params)
        query_params_for_finalize = f'{query_params}&nb_rows={nb_rows_to_be_inserted}'
        finalize_url = '{}/ingestion/finalize?{}'.format(self._k.kawa_api_url, query_params_for_finalize)
        finalize_for_failure_url = '{}/ingestion/stop-with-failure?{}'.format(self._k.kawa_api_url, query_params)

        # Check that all date and date time indicators are numbers in the data frame
        temporal_indicators = [i for i in indicators if i.get('type') == 'date_time']
        for temporal_indicator in temporal_indicators:
            column_name = temporal_indicator.get('indicatorId')
            if column_name in df.columns:
                column_type_name = str(df[column_name].dtype)
                if re.match(r'^datetime64', column_type_name):
                    try:
                        if self._use_arrow:
                            timestamp_column = self._arrow_table[column_name]
                            nanos = pc.cast(timestamp_column, pa.int64())
                            millis = pc.divide(nanos, 1000000)
                            self._arrow_table = self._arrow_table.set_column(
                                self._arrow_table.schema.get_field_index(column_name),
                                column_name,
                                millis
                            )
                        else:
                            df[column_name] = df[column_name].map(lambda x: int(x.timestamp() * 1000) if x else None)
                    except ValueError as e:
                        # todo:
                        #  fix 'ValueError: The truth value of a DatetimeIndex is ambiguous.
                        #  Use a.empty, a.bool(), a.item(), a.any() or a.all()'
                        traceback.print_exc()

        # Replace None/NaN with empty strings in text columns
        text_column_names = []
        text_indicators = [i for i in indicators if i.get('type') == 'text']
        for text_indicator in text_indicators:
            column_name = text_indicator.get('indicatorId')
            if column_name in df.columns:
                text_column_names.append(column_name)

        df.fillna(value={v: '' for v in text_column_names}, inplace=True)

        # Call prepare data that will check if we can start loading and give us the offset for automatic index
        prepare_data = self._k.post(url=prepare_url, data={})

        if not prepare_data.get('canRunLoading'):
            raise Exception(
                'We cannot start ingestion due to: ' + prepare_data.get('raisonItCannotStart', 'No reason given'))

        parquet_directory = '{}/{}'.format(self._k.tmp_files_directory, str(uuid.uuid4()))
        try:
            auto_increment_indicator = [i for i in indicators if
                                        i.get('storageConfig', {}).get('automaticUniqueValue', False)]

            if len(auto_increment_indicator) == 1:
                if 'offsetToApplyToAutoIncrementIndex' not in prepare_data:
                    self._k.post(url=finalize_for_failure_url, data={})
                    raise Exception('The offset for to the auto_increment_index was not present in the answer from '
                                    'backend. Cannot continue')

                auto_increment_indicator_id = auto_increment_indicator[0].get('indicatorId')
                offset = prepare_data.get('offsetToApplyToAutoIncrementIndex')

                if self._use_arrow:
                    if auto_increment_indicator_id in self._arrow_table.schema.names:
                        self._arrow_table = self._arrow_table.remove_column(
                            self._arrow_table.schema.get_field_index(auto_increment_indicator_id))

                    index_vector = pa.array(range(offset + 1, offset + 1 + self._arrow_table.num_rows))
                    self._arrow_table = self._arrow_table.append_column(
                        auto_increment_indicator_id,
                        index_vector)
                else:
                    index_vector = numpy.arange(0, df.shape[0], 1)
                    df[auto_increment_indicator_id] = index_vector + offset + 1

            # Check that all the indicators are present in the data frame, otherwise create empty columns
            for indicator in indicators:
                indicator_id = indicator.get('indicatorId')
                if self._use_arrow and indicator_id not in self._arrow_table.schema.names:
                    default_value = self._empty_value_for_indicator(indicator)
                    new_column = pa.array([default_value] * self._arrow_table.num_rows)
                    self._arrow_table = self._arrow_table.append_column(indicator_id, new_column)

                elif indicator_id not in df.columns:
                    default_value = self._empty_value_for_indicator(indicator)
                    df[indicator_id] = default_value

            if self._use_arrow:
                start = time.time()
                print('> Starting loading arrow stream')
                self._loading_thread_arrow(
                    ingestion_url=ingest_url,
                    arrow_table=self._arrow_table
                )
                num_rows = self._arrow_table.num_rows
                end = time.time()
                print('> {} rows were imported in {}s'.format(num_rows, end - start))
                return datasource

            # Add a partition column to split up the frames into multiple parquet files
            if not parquet_file_list:
                partition_cols = []
                nb_partitions = max(1, nb_threads)
                if nb_partitions > 1:
                    index_vector = numpy.arange(0, df.shape[0], 1)
                    df['__partition__'] = (index_vector + 1) % nb_partitions
                    partition_cols.append('__partition__')

                if show_progress:
                    print('> Exporting the dataframe into {} parquet file{}'.format(nb_partitions,
                                                                                    's' if nb_partitions > 1 else ''))
                os.makedirs(parquet_directory, exist_ok=True)
                df.to_parquet(partition_cols=partition_cols, path=parquet_directory + '/', compression='gzip')

            start = time.time()

            if parquet_file_list:
                parquet_files = parquet_file_list
            else:
                parquet_files = glob.glob(
                    pathname='{}/**/*.parquet'.format(parquet_directory),
                    recursive=True,
                )

            if show_progress:
                print('> Starting {} loading threads'.format(nb_threads))

            with concurrent.futures.ThreadPoolExecutor(max_workers=nb_threads) as executor:
                futures = [
                    executor.submit(self._loading_thread, ingest_url, parquet_file)
                    for parquet_file
                    in parquet_files
                ]
                concurrent.futures.wait(futures)

            end = time.time()
            if show_progress:
                print('> {} rows were imported in {}ms'.format(df.shape[0], end - start))

        except Exception as e:
            self._k.post(url=finalize_for_failure_url, data={})
            raise e

        finally:
            if os.path.isdir(parquet_directory):
                shutil.rmtree(parquet_directory)

            self._k.post(url=finalize_url, data={})
            if show_progress:
                print('> Import was successfully finalized')

        return datasource

    def _loading_thread(self, ingestion_url, parquet_file):
        self._k.post_binary_file(filename=parquet_file, url=ingestion_url)

    def _loading_thread_arrow(self, ingestion_url, arrow_table: pa.Table):

        def stream_batches():
            output_stream = io.BytesIO()
            schema = arrow_table.schema
            writer = ipc.RecordBatchStreamWriter(output_stream, schema)
            chunk_size = 10000
            for start in range(0, len(arrow_table), chunk_size):
                chunk = arrow_table.slice(start, chunk_size)
                writer.write_table(chunk)
                output_stream.seek(0)
                yield output_stream.getvalue()
                output_stream.truncate(0)
                output_stream.seek(0)

            writer.close()

        self._k.post_stream(
            stream=stream_batches(),
            url=ingestion_url
        )

    @staticmethod
    def _empty_value_for_indicator(indicator):
        indicator_type = indicator.get('type')
        if indicator_type == 'text':
            return ''
        if indicator_type == 'date':
            return date(1970, 1, 1)
        if indicator_type == 'date_time':
            return 0
        if indicator_type.startswith('list('):
            raise 'Does not support omitting lists'
        return None



from ...client.data_generator import DataGenerator
from datetime import date
import uuid
import numpy

from ..kawa_base_e2e_test import KawaBaseTest


class TestDataGenerator(KawaBaseTest):

    def test_that_data_is_properly_generated(self):
        # setup
        sheet_name = 'generated_' + str(uuid.uuid4())
        generator = DataGenerator(
            kawa_client=self.kawa,
            sheet_and_datasource_name=sheet_name,
            number_of_rows=10,
            number_of_measures=1,
            number_of_dimensions=1
        )

        # when
        generator.generate_data()

        # then
        computed_df = (self.kawa
                       .sheet(sheet_name=sheet_name)
                       .select(self.kawa.cols())
                       .compute())

        self.assertEqual(computed_df.shape[0], 10)
        self.assertGreaterEqual(computed_df.shape[1], 13)

        self.assertListEqual(list(computed_df['f_id']), ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9'])

        self.assertTrue(type(computed_df['f_id'][0]) is str)
        self.assertTrue(type(computed_df['dim_0'][0]) is str)
        self.assertTrue(type(computed_df['f_dimension_10'][0]) is str)
        self.assertTrue(type(computed_df['f_dimension_100'][0]) is str)
        self.assertTrue(type(computed_df['f_dimension_1000'][0]) is str)
        self.assertTrue(type(computed_df['f_pseudo_id_2'][0]) is str)
        self.assertTrue(type(computed_df['f_pseudo_id_3'][0]) is str)
        self.assertTrue(type(computed_df['f_pseudo_id_4'][0]) is str)
        self.assertTrue(type(computed_df['f_pseudo_id_5'][0]) is str)
        self.assertTrue(type(computed_df['f_pseudo_id_6'][0]) is str)
        self.assertTrue(type(computed_df['f_text'][0]) is str)
        self.assertTrue(type(computed_df['f_date'][0]) is date)
        self.assertTrue(type(computed_df['meas_0'][0]) is numpy.int64)



import tempfile
import uuid
from datetime import date, datetime

import pandas as pd
import pytz

from ...kawa_base_e2e_test import KawaBaseTest
from ....client.computation_nodes import add, val, col


def utc(year, month, day, hour, minute, second):
    return datetime(year, month, day, hour, minute, second, tzinfo=pytz.UTC)


class TestComputedColumns(KawaBaseTest):

    @classmethod
    def setUpClass(cls):
        unique_id = 'resource_{}'.format(uuid.uuid4())
        print(tempfile.gettempdir())
        KawaBaseTest.setUpClass()
        cls.data, cls.df = cls.df_for_tests()
        cls.sheet_name = unique_id
        loader = cls.kawa.new_data_loader(df=cls.df.copy(deep=True), datasource_name=unique_id)
        loader.create_datasource()
        loader.load_data(reset_before_insert=True, create_sheet=True)

    def test_select_computed_column(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .select(add(val(5), col('integer')).alias('5 plus integer column'))
                       .compute())

        # then
        self.assertListEqual(list(computed_df['5 plus integer column']), [i + 5 for i in self.data['integer']])

    def test_select_formula_without_aliases(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .select(add(val(5), col('integer')))
                       .compute())

        # then
        self.assertListEqual(list(computed_df['col0']), [i + 5 for i in self.data['integer']])

    def test_select_formula_and_columns(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .select(add(val(1), val(2)).alias('1 plus 2'), 'text')
                       .compute())

        # then
        self.assertListEqual(list(computed_df.columns), ['1 plus 2', 'text'])

    @staticmethod
    def df_for_tests():
        data = {
            'date': [
                date(2023, 1, 1),
                date(2023, 1, 2),
                date(2023, 1, 1),
                date(2023, 1, 2),
                date(2023, 1, 2)
            ],
            'date_time': [
                utc(2023, 1, 1, 23, 23, 2),
                utc(2023, 1, 1, 23, 23, 4),
                utc(2023, 1, 1, 23, 23, 2),
                utc(2023, 1, 1, 23, 23, 2),
                utc(2023, 1, 1, 23, 23, 2),
            ],
            'text': ['value11', 'value12', 'value21', 'value22', 'value23'],
            'integer': [1, 2, 3, 4, 5],
            'decimal': [1.1, 2.1, 3.1, 4.1, 5.1],
            'boolean': [True, False, True, True, False],
        }

        data_for_df = []
        for i in range(5):
            data_for_df.append({
                'date': data['date'][i],
                'date_time': data['date_time'][i],
                'boolean': data['boolean'][i],
                'text': data['text'][i],
                'integer': data['integer'][i],
                'decimal': data['decimal'][i],
            })

        return data, pd.DataFrame(data_for_df)



import os
import unittest

from ..client.commands import SourceControlType
from ..client.kawa_client import KawaClient as kawa
from dotenv import load_dotenv

load_dotenv()

kawa_url = os.environ['E2E_KAWA_URL']
kawa_api_key_user_1 = os.environ['E2E_KAWA_API_KEY_USER_1']
kawa_api_key_user_2 = os.environ['E2E_KAWA_API_KEY_USER_2']
workspace = os.getenv('E2E_KAWA_WORKSPACE', 1)
git_hub_key = os.getenv('E2E_KAWA_GIT_HUB_KEY')
git_lab_key = os.getenv('E2E_KAWA_GIT_LAB_KEY')


class KawaBaseTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        print('Setup tests in workspace {} on url {}'.format(workspace, kawa_url))
        k = kawa(kawa_api_url=kawa_url)
        k_second_user = kawa(kawa_api_url=kawa_url)

        k.set_api_key(api_key=kawa_api_key_user_1)
        k.set_active_workspace_id(workspace_id=workspace)
        cls.kawa = k
        cls.commands = k.commands
        cls.workspace_id = workspace

        k_second_user.set_api_key(api_key=kawa_api_key_user_2)
        k_second_user.set_active_workspace_id(workspace_id=workspace)
        cls.kawa_second_user = k_second_user
        cls.commands_second_user = k_second_user.commands

        # GITHUB
        # k.commands.set_source_control_config_to_workspace(workspace,
        #                                                   'https://api.github.com',
        #                                                   git_hub_key,
        #                                                   'greghass/test-kawa-packages',
        #                                                   'main',
        #                                                   SourceControlType.GIT_HUB)

        # GITLAB
        # k.commands.set_source_control_config_to_workspace(workspace,
        #                                                   'https://gitlab.com',
        #                                                   git_lab_key,
        #                                                   'greghass/test-kawa-packages',
        #                                                   'main',
        #                                                   SourceControlType.GIT_LAB)



from ..kawa_base_e2e_test import KawaBaseTest
import pandas as pd
import uuid
import tempfile


class TestSheetProfiling(KawaBaseTest):
    cities_and_countries = pd.DataFrame([
        {'id': 'a', 'country': 'FR', 'measure': 2},
        {'id': 'b', 'country': 'FR', 'city': 'Lyon', 'measure': 2},
        {'id': 'c', 'country': 'FR', 'city': 'Lille', 'measure': 2},
        {'id': 'd', 'country': 'UK', 'city': 'London', 'measure': 2},
        {'id': 'e', 'country': 'UK', 'city': 'Cardiff', 'measure': 2},
        {'id': 'g', 'country': 'UK', 'city': 'Edinburgh', 'measure': 2},
        {'id': 'h', 'country': 'UK', 'city': 'Belfast', 'measure': 2},
        {'id': 'i', 'country': 'BE', 'city': 'Brussels', 'measure': 2},
        {'id': 'j', 'country': 'BE', 'city': 'Bruges', 'measure': 2},
    ])

    @classmethod
    def setUpClass(cls):
        unique_id = 'resource_{}'.format(uuid.uuid4())
        print(tempfile.gettempdir())
        KawaBaseTest.setUpClass()
        cls.sheet_name = unique_id

    def test_that_sheet_profiling_works_on_a_given_column(self):
        # setup
        df = self.cities_and_countries
        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader = self.kawa.new_data_loader(df=df, datasource_name=unique_id)
        loader.create_datasource(primary_keys=['id'])
        loader.load_data(reset_before_insert=True, create_sheet=True)

        # when
        profile = (self.kawa
                   .sheet(sheet_name=unique_id)
                   .profile(column_name='city'))

        # then
        self.assertEqual(len(profile), 1)
        self.assertDictEqual(profile['city'], {
            'COUNT': 8,
            'COUNT_UNIQUE': 8,
            'COUNT_EMPTY': 1,
        })

    def test_that_sheet_profiling_works_on_all_columns(self):
        # setup
        df = self.cities_and_countries
        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader = self.kawa.new_data_loader(df=df, datasource_name=unique_id)
        loader.create_datasource(primary_keys=['id'])
        loader.load_data(reset_before_insert=True, create_sheet=True)

        # when
        profile = self.kawa.sheet(sheet_name=unique_id).profile()

        # then
        self.assertEqual(len(profile), 4)
        self.assertDictEqual(profile['city'], {
            'COUNT': 8,
            'COUNT_UNIQUE': 8,
            'COUNT_EMPTY': 1,
        })
        self.assertDictEqual(profile['country'], {
            'COUNT': 9,
            'COUNT_UNIQUE': 3,
            'COUNT_EMPTY': 0,
        })
        self.assertDictEqual(profile['id'], {
            'COUNT': 9,
            'COUNT_UNIQUE': 9,
            'COUNT_EMPTY': 0,
        })
        self.assertDictEqual(profile['measure'], {
            'STD_DEV_POP': 0,
            'MIN': 2,
            'HIGHEST_DECILE': 2,
            'AVERAGE': 2,
            'MAX': 2,
            'LOWEST_QUARTILE': 2,
            'LOWEST_DECILE': 2,
            'SUM': 18,
            'MEDIAN': 2,
            'VAR_POP': 0,
            'HIGHEST_QUARTILE': 2,
        })
